@@ -7,7 +7,7 @@
 # Contents:
 #   General-purpose classes and functions.
 #
-# Copyright (c) 2000, 2001, 2002 by CodeSourcery, LLC.  All rights reserved. 
+# Copyright (c) 2000, 2001, 2002, 2003 by CodeSourcery, LLC.  All rights reserved. 
 #
 # For license terms see the file COPYING.
 #
@@ -18,17 +18,19 @@
 ########################################################################
 
 import base64
+from   calendar import timegm
 import ConfigParser
 import cPickle
 import cStringIO
 import dircache
 import gzip
 import imp
+import lock
 import operator
 import os
 import os.path
+import qm
 import quopri
-import random
 import re
 import socket
 import string
@@ -37,12 +39,14 @@ import tempfile
 import time
 import traceback
 import types
-
+if sys.platform != "win32":
+    import fcntl
+    
 ########################################################################
 # program name
 ########################################################################
 
-program_name = "??"
+program_name = None
 """The name of the application program."""
 
 ########################################################################
@@ -55,18 +59,20 @@ class Empty:
     pass
 
 
+
 class QMException(Exception):
     """An exception generated directly by QM.
 
     All exceptions thrown by QM should be derived from this class."""
-    
-    pass
 
+    def __init__(self, message):
+        """Construct a new 'QMException'.
 
-class MethodShouldBeOverriddenError(QMException):
-    """Indicates a method was called that should have been overridden."""
+        'message' -- A string describing the cause of the message as
+        structured text.  If this exception is not handled, the
+        'message' will be displayed as an error message."""
 
-    pass
+        Exception.__init__(self, message)
 
 
 
@@ -96,142 +102,36 @@ class UserError(QMException):
 
 
 
+class PythonException(QMException):
+    """A 'PythonException' is a wrapper around a Python exception.
+
+    A 'PythonException' is a 'QMException' and, as such, can be
+    processed by the QM error-handling routines.  However, the raw
+    Python exception which triggered this exception can be obtained by
+    using the 'exc_type' and 'exc_value' attributes of this
+    exception."""
+
+    def __init__(self, message, exc_type, exc_value):
+        """Construct a new 'PythonException'.
+
+        'message' -- A string describing the cause of the message as
+        structured text.  If this exception is not handled, the
+        'message' will be displayed as an error message.
+
+        'exc_type' -- The type of the Python exception.
+
+        'exc_value' -- The value of the Python exception."""
+        
+        QMException.__init__(self, message)
+
+        self.exc_type = exc_type
+        self.exc_value = exc_value
+    
 ########################################################################
 # classes
 ########################################################################
 
-class FileSystemMutex:
-    """A mutual exclusion lock residing in the file system."""
-
-    retry_interval = 0.1
-    """The interval, in seconds, at which to retry a lock."""
-
-
-    def __init__(self, path):
-        "Create a new mutex at 'path'.  No lock is aquired."""
-
-        self.__path = path
-        self.__pid_filename = os.path.join(path, "pid")
-
-        # Perform sanity check.  If 'path' exists, it should be a
-        # lock held by another instance of this class.
-        if os.path.exists(self.__path):
-            if not os.path.isdir(self.__path):
-                raise RuntimeError, "path exists and isn't a directory"
-
-        self.__locked = 0
-
-
-    def GetPath(self):
-        """Return the path of the lock."""
-
-        return self.__path
-
-
-    def Lock(self, timeout=None):
-        """Aquire a lock.  If the mutex is already held, block.
-
-        'timeout' -- If 'None', the lock operation blocks
-        indefinitely, until the mutex is available.  If not 'None',
-        this is a timeout, in seconds, after which to give up.  If
-        zero, this function returns immediately if a lock cannot be
-        acquired.
-
-        raises -- 'MutexError' if there is a problem with the mutex.
-
-        raises -- 'MutexLockError' if a lock was not acquired."""
-
-        # Don't allow double locks.
-        if self.__locked:
-            return
-
-        start_time = time.time()
-        while 1:
-            parent_dir = os.path.dirname(self.__path)
-            # Make sure the parent directory exists.
-            if not os.path.isdir(parent_dir):
-                raise MutexError, \
-                      "parent directory %s doesn't exist" % parent_dir
-            # Check if the directory exists.
-            if not os.path.isdir(self.__path):
-                # If not, attempt to create it.
-                try:
-                    os.mkdir(self.__path)
-                    # Creation succeeded; the lock is ours.
-                    self.__locked = 1
-                    # Write our pid into a file in the directory.
-                    pid_file = open(self.__pid_filename, "w")
-                    pid_file.write("%d\n" % os.getpid())
-                    pid_file.close()
-                    return
-                except os.error:
-                    # Couldn't lock.
-                    pass
-            # Time out yet?
-            if timeout != None \
-               and (time.time() - start_time) >= timeout:
-                # Timed out.  Raise an exception.
-                raise MutexLockError, \
-                      "lock on %s timed out" % self.__path
-            # Sleep for a while before trying again.
-            time.sleep(self.retry_interval)
-
-        
-    def Unlock(self):
-        """Release a lock."""
-
-        # Don't try to unlock if we're not already locked.
-        if not self.__locked:
-            return
-
-        # Sanity checks.
-        assert os.path.isdir(self.__path)
-        assert os.path.isfile(self.__pid_filename)
-
-        # Unlock.
-        os.unlink(self.__pid_filename)
-        os.rmdir(self.__path)
-        self.__locked = 0
-
-
-    def IsLocked(self):
-        """Return true if a lock is held on the mutex."""
-
-        return self.__locked
-        
-        
-
-class MapReplacer:
-    """A callable object to replace text according to a map."""
-
-    def __init__(self, replacements):
-        """Generate a function to replace text according to a map.
-
-        'replacements' -- A mapping of replacements.  For each element,
-        the key is the text to match, and the corresponding value is the
-        replacements for that text."""
-
-        # Construct a regular expression that matches any of the
-        # replacements.
-        keys = map(re.escape, replacements.keys())
-        regex = "(" + string.join(keys, "|") + ")"
-        self.__regex = re.compile(regex)
-        # The replacement function.  It simply looks up the replacement text
-        # in 'replacements'.
-        self.__substitution = lambda match, replacements=replacements: \
-                              replacements[match.group(0)]
-
-
-    def __call__(self, text):
-        """Perform replacements in 'text'.
-
-        returns -- The replaced text."""
-
-        return self.__regex.sub(self.__substitution, text)
-
-
-
-class RcConfiguration:
+class RcConfiguration(ConfigParser.ConfigParser):
     """Interface object to QM configuration files.
 
     Configuration files are in the format parsed by the standard
@@ -244,16 +144,22 @@ class RcConfiguration:
     def __init__(self):
         """Create a new configuration instance."""
 
-        self.__parser = None
+        ConfigParser.ConfigParser.__init__(self)
+        if os.environ.has_key("HOME"):
+            home_directory = os.environ["HOME"]
+            rc_file = os.path.join(home_directory, self.user_rc_file_name)
+            # Note that it's OK to call 'read' even if the file doesn't
+            # exist.  In that, case the parser simply will not
+            # accumulate any data.
+            self.read(rc_file)
 
 
     def Load(self, section):
         """Load configuration.
 
         'section' -- The configuration section from which subsequent
-        varaibles are loaded."""
+        variables are loaded."""
 
-        self.__parser = self.__Load()
         self.__section = section
 
 
@@ -271,15 +177,6 @@ class RcConfiguration:
 
         precondition -- The RC configuration must be loaded."""
 
-        if self.__parser is None:
-            # No RC file was ever loaded.  Print a warning the first
-            # time. 
-            if not hasattr(self, "no_rc_loaded_warning"):
-                sys.stderr.write("Warning: No RC configuration file loaded.\n")
-                self.no_rc_loaded_warning = 1
-            # Use the default.             
-            return default
-        
         # Use the previously-specified default section, if one wasn't
         # specified explicitly.
         if section is None:
@@ -287,7 +184,7 @@ class RcConfiguration:
 
         try:
             # Try to get the requested option.
-            return self.__parser.get(section, option)
+            return self.get(section, option)
         except ConfigParser.NoSectionError:
             # Couldn't find the section.
             return default
@@ -309,7 +206,7 @@ class RcConfiguration:
         if section is None:
             section = self.__section
         try:
-            options = self.__parser.options(section)
+            options = self.options(section)
         except ConfigParser.NoSectionError:
             # Couldn't find the section.
             return []
@@ -320,160 +217,30 @@ class RcConfiguration:
                 options.remove("__name__")
             return options
 
-
-    def __Load(self):
-        """Load the configuration from the appropriate places."""
-
-	# Create a parser.
-        parser = ConfigParser.ConfigParser()
-
-        # Construct the path to the user's rc file.
-        if os.environ.has_key("HOME"):
-	        home_directory = os.environ["HOME"]
-	        rc_file = os.path.join(home_directory, self.user_rc_file_name)
-	        # Note that it's OK to call 'read' even if the file doesn't
-		# exist.  In that, case the parser simply will not accumulate
-		# any data.
-	        parser.read(rc_file)
-	else:	
-		# If we cannot find the user's home directory, do not
-		# even try to read the configuration.
-		pass
-
-        return parser
-
-
-
-class OrderedMap:
-    """A map that preserves the order of elements inserted into it.
-
-    An 'OrderedMap' object behaves in many ways like a map.
-
-      - It supports efficient lookup of value by key.
-
-      - It supports 'keys' and 'values' operations efficiently.  Both
-        these lists present their elements in the order they were
-        inserted into the map.
-
-      - It does not support the 'items' operation.
-
-      - It takes more space than an ordinary map.
-
-      - It does not support deleting of elements.  The value of an
-        exisiting key may be replaced efficiently, though; this does not
-        change the position of the key in the order.
-    """
-
-    # The '__keys' and '__values' lists store keys and values at
-    # corresponding indices.  The '__key_map' map associates keys with
-    # indices in these two lists.
-
-    def __init__(self):
-        self.__keys = []
-        self.__values = []
-        self.__key_map = {}
-
-
-    def __getitem__(self, key):
-        return self.__values[self.__key_map[key]]
-
-
-    def __setitem__(self, key, value):
-        try:
-            index = self.__key_map[key]
-        except KeyError:
-            index = len(self.__keys)
-            self.__keys.append(key)
-            self.__values.append(value)
-            self.__key_map[key] = index
-        else:
-            self.__values[index] = value
-
-
-    def __len__(self):
-        return len(self.__keys)
     
-
-    def has_key(self, key):
-        return self.__key_map.has_key(key)
-
-
-    def keys(self):
-        return self.__keys
-
-
-    def values(self):
-        return self.__values
-
-
-    def get(self, key, default=None):
-        try:
-            index = self.__key_map[key]
-        except KeyError:
-            return default
-        else:
-            return self.__values[index]
-
-
-    def clear(self):
-        self.__keys = []
-        self.__values = []
-        self.__key_map.clear()
-
-
-    def copy(self):
-        result = OrderedMap()
-        result.__keys = self.__keys[:]
-        result.__values = self.__values[:]
-        result.__key_map = self.__key_map.copy()
-        return result
-
-
-    def update(self, map):
-        for key, value in map.items():
-            self[key] = value
-            
-
-    def __str__(self):
-        pairs = map(lambda i, k=self.__keys, v=self.__values: 
-                        repr(k[i]) + ": " + repr(v[i]),
-                    range(0, len(self.__keys)))
-        return "OrderedMap{%s}" % string.join(pairs, ", ")
-
-
-    def AsMap(self):
-        """Return an ordinary map object."""
-
-        result = {}
-        for i in range(0, len(self.__keys)):
-            result[self.__keys[i]] = self.__values[i]
-        return result
-
-
-
 ########################################################################
-# functions
+# Functions
 ########################################################################
 
 def get_lib_directory(*components):
-    """Return the absolute path to the top QM Python directory."""
+    """Return the path to a file in the QM library directory."""
 
-    path = os.environ["QM_LIB_PATH"]
-    return apply(os.path.join, (path, ) + components)
+    return os.path.join(qm.prefix, qm.lib_dir, *components)
 
 
 def get_share_directory(*components):
-    """Return the path to the directory containing QM data files."""
+    """Return the path to a file in the QM data file directory."""
 
-    path = os.environ["QM_SHARE_PATH"]
-    return apply(os.path.join, (path, ) + components)
+    return os.path.join(qm.prefix, qm.data_dir, *components)
 
 
 def get_doc_directory(*components):
-    """Return the path to the directory containing QM documentation files."""
+    """Return a path to a file in the QM documentation file directory."""
 
-    path = os.environ["QM_DOC_PATH"]
-    return apply(os.path.join, (path, ) + components)
+    if not is_installed:
+        return os.path.join(qm.prefix, "qm", *components)
+    else:
+        return os.path.join(get_share_directory("doc"), *components)
 
 
 def format_exception(exc_info):
@@ -499,7 +266,7 @@ def format_traceback(exc_info):
 
     returns -- A string containing a the formatted traceback."""
 
-    return string.join(traceback.format_tb(exc_info[2]))
+    return string.join(traceback.format_tb(exc_info[2]), "\n")
 
 
 def format_byte_count(bytes):
@@ -522,64 +289,25 @@ def format_byte_count(bytes):
     return "%d bytes" % bytes
 
 
-def rmdir_recursively(path):
-    """Remove the directory at 'path' and everything under it."""
-
-    # Remove everything in the directory.
-    for entry in dircache.listdir(path):
-        entry_path = os.path.join(path, entry)
-        if os.path.isdir(entry_path):
-            rmdir_recursively(entry_path)
-        else:
-            os.unlink(entry_path)
-    # Remove the directory itself.
-    os.rmdir(path)
-
-
-def replace_by_map(text, replacements):
-    """Perform multiple replacements.
-
-    'text' -- The text in which to make the replacements.
-
-    'replacements' -- A mapping of replacements.  For each element, the
-    key is the text to match, and the corresponding value is the
-    replacements for that text.
-
-    returns -- The replaced text."""
-
-    return MapReplacer(replacements)(text)
-
-
-def invert_map(m):
-    """Return the inverse of 'm'.
-
-    'm' -- A map object.
-
-    returns -- A map of values of 'm' to corresponding keys.  If a value
-    of 'm' is associated with more than one key, it is mapped to one of
-    these keys in the result, but which one is undefined."""
-
-    result = {}
-    for key, value in m.items():
-        result[value] = key
-    return result
-
-
 def convert_from_dos_text(text):
     """Replace CRLF with LF in 'text'."""
 
     return string.replace(text, "\r\n", "\n")
 
+__load_module_lock = lock.RLock()
+"""A lock used by load_module."""
 
-def load_module(name, path=sys.path):
+def load_module(name, search_path=sys.path, load_path=sys.path):
     """Load a Python module.
 
     'name' -- The fully-qualified name of the module to load, for
     instance 'package.subpackage.module'.
 
-    'path' -- A sequence of directory paths in which to search for the
-    module, analogous to 'PYTHONPATH'.
+    'search_path' -- A sequence of directories.  These directories are
+    searched to find the module.
 
+    'load_path' -- The setting of 'sys.path' when the module is loaded.
+    
     returns -- A module object.
 
     raises -- 'ImportError' if the module cannot be found."""
@@ -589,60 +317,69 @@ def load_module(name, path=sys.path):
     # the 'knee' package included unofficially in the standard Python
     # library. 
 
-    # Is the module already loaded?
+    # In order to avoid getting incomplete modules, grab a lock here.
+    # Use a recursive lock so that deadlock does not occur if the loaded
+    # module loads more modules.
+    __load_module_lock.acquire()
     try:
-        module = sys.modules[name]
-        # It's listed; is there a module instance there?
-        if module is not None:
+        # Is the module already loaded?
+        module = sys.modules.get(name)
+        if module:
             return module
-        # Nope; go on loading.
-    except KeyError:
-        # No; that's OK.
-        pass
 
-    # The module may be in a package.  Split the module path into
-    # components. 
-    components = string.split(name, ".")
-    if len(components) > 1:
-        # The module is in a package.  Construct the name of the
-        # containing package.
-        parent_package = string.join(components[:-1], ".")
-        # Load the containing package.
-        package = load_module(parent_package, path)
-        # Look for the module in the parent package.
-        path = package.__path__
-    else:
-        # No containing package.
-        package = None
-    # The name of the module itself is the last component of the module
-    # path.  
-    module_name = components[-1]
-    # Locate the module.
-    file, file_name, description = imp.find_module(module_name, path)
-    # Find the module.
-    try:
-        # While loading the module, add 'path' to Python's module path,
-        # so that if the module references other modules, e.g. in the
-        # same directory, Python can find them.  But remember the old
-        # path so we can restore it afterwards.
-        old_python_path = sys.path[:]
-        sys.path = sys.path + path
-        # Load the module.
-        module = imp.load_module(name, file, file_name, description)
-        # Restore the old path.
-        sys.path = old_python_path
-        # Loaded successfully.  If it's contained in a package, put it
-        # into that package's name space.
-        if package is not None:
-            setattr(package, module_name, module)
-        return module
+        # The module may be in a package.  Split the module path into
+        # components. 
+        components = string.split(name, ".")
+        if len(components) > 1:
+            # The module is in a package.  Construct the name of the
+            # containing package.
+            parent_package = string.join(components[:-1], ".")
+            # Load the containing package.
+            package = load_module(parent_package, search_path, load_path)
+            # Look for the module in the parent package.
+            search_path = package.__path__
+        else:
+            # No containing package.
+            package = None
+        # The name of the module itself is the last component of the module
+        # path.  
+        module_name = components[-1]
+        # Locate the module.
+        file, file_name, description = imp.find_module(module_name,
+                                                       search_path)
+        # Find the module.
+        try:
+            # While loading the module, add 'path' to Python's module path,
+            # so that if the module references other modules, e.g. in the
+            # same directory, Python can find them.  But remember the old
+            # path so we can restore it afterwards.
+            old_python_path = sys.path[:]
+            sys.path = load_path + sys.path
+            # Load the module.
+            try:
+                module = imp.load_module(name, file, file_name, description)
+            except:
+                # Don't leave a broken module object in sys.modules.
+                if sys.modules.has_key(name):
+                    del sys.modules[name]
+                raise
+            # Restore the old path.
+            sys.path = old_python_path
+            # Loaded successfully.  If it's contained in a package, put it
+            # into that package's name space.
+            if package is not None:
+                setattr(package, module_name, module)
+            return module
+        finally:
+            # Close the module file, if one was opened.
+            if file is not None:
+                file.close()
     finally:
-        # Close the module file, if one was opened.
-        if file is not None:
-            file.close()
+        # Release the lock.
+        __load_module_lock.release()
+
         
-        
-def load_class(name, path=sys.path):
+def load_class(name, search_path = sys.path, load_path = sys.path):
     """Load a Python class.
 
     'name' -- The fully-qualified (including package and module names)
@@ -650,8 +387,10 @@ def load_class(name, path=sys.path):
     class must be at the top level of the module's namespace, i.e. not
     nested in another class.
 
-    'path' -- A sequence of directory paths in which to search for the
-    containing module, analogous to 'sys.path'.
+    'search_path' -- A sequence of directories.  These directories are
+    searched to find the module.
+
+    'load_path' -- The setting of 'sys.path' when the module is loaded.
 
     returns -- A class object.
 
@@ -672,11 +411,16 @@ def load_class(name, path=sys.path):
     # The last element is the name of the class.
     class_name = components[-1]
     # Load the containing module.
-    module = load_module(module_name, path)
-    # Exctract the requested class.
+    module = load_module(module_name, search_path, load_path)
+    # Extract the requested class.
     try:
         klass = module.__dict__[class_name]
-        if not isinstance(klass, types.ClassType):
+        # Check to see the KLASS really is a class.  Python 2.2's
+        # "new-style" classes are not instances of types.ClassType so we
+        # must check two conditions: one for old-style and one for
+        # new-style classes.
+        if (not isinstance(klass, types.ClassType)
+            and not issubclass(klass, object)):
             # There's something by that name, but it's not a class
             raise QMException, "%s is not a class" % name
         return klass
@@ -763,18 +507,20 @@ def decode_data_from_text(data, encoding):
         raise ValueError, "unknown encoding %s" % encoding
 
 
-def open_temporary_file_fd():
+def open_temporary_file_fd(suffix = ""):
     """Create and open a temporary file.
 
+    'suffix' -- The last part of the temporary file name, as for
+    Python's 'mktemp' function.
+    
     The file is open for reading and writing.  The caller is responsible
     for deleting the file when finished with it.
 
     returns -- A pair '(file_name, file_descriptor)' for the temporary
     file."""
 
-    # FIXME: Security.
+    file_name = tempfile.mktemp(suffix)
 
-    file_name = tempfile.mktemp()
     try:
         # Attempt to open the file.
         fd = os.open(file_name,
@@ -782,115 +528,50 @@ def open_temporary_file_fd():
                      0600)
     except:
         exc_info = sys.exc_info()
-        raise RuntimeError, qm.error("temp file error",
-                                     file_name=file_name,
-                                     exc_class=str(exc_info[0]),
-                                     exc_arg=str(exc_info[1]))
+        raise QMException, \
+              qm.error("temp file error",
+                       file_name=file_name,
+                       exc_class=str(exc_info[0]),
+                       exc_arg=str(exc_info[1]))
     return (file_name, fd)
 
 
-def open_temporary_file():
+def open_temporary_file(mode = "w+b", suffix = ""):
     """Create and open a temporary file.
 
+    'mode' -- The mode argument to pass to 'fopen'.
+    
+    'suffix' -- The last part of the temporary file name, as for
+    Python's 'mktemp' function.
+    
     Like 'open_temporary_file_fd', except that the second element of the
     return value is a file object."""
 
-    file_name, fd = open_temporary_file_fd()
-    return (file_name, os.fdopen(fd, "w+b"))
+    # This function can just use NamedTemporaryFile when QMTest requires
+    # Python 2.3.
+    file_name, fd = open_temporary_file_fd(suffix)
+    return (file_name, os.fdopen(fd, mode))
 
 
-def make_temporary_directory():
-    """Create a temporary directory.
-
-    returns -- The path to the temporary directory."""
-
-    dir_path = tempfile.mktemp()
-    try:
-        os.mkdir(dir_path, 0700)
-    except:
-        exc_info = sys.exc_info()
-        raise RuntimeError, qm.error("temp dir error",
-                                     dir_path=dir_path,
-                                     exc_class=str(exc_info[0]),
-                                     exc_arg=str(exc_info[1]))
-    return dir_path
-
-
-def find_program_in_path(program_name, path=None):
-    """Attempt to locate a program in an execution path.
-
-    'program_name' -- The name of the program to run.
-
-    'path' -- A string encoding a list of directories (in a
-    system-specifiec form) in which to look for the program.  If
-    'None', the 'PATH' environment variable is used.
-
-    returns -- The path to the program, or 'None' if it was not
-    found."""
-
-    # If no path was specified, use the PATH environment variable.
-    if path is None:
-        path = os.environ["PATH"]
-    # Split the path into directories.
-    path_separator = os.pathsep
-    directories = string.split(path, path_separator)
-    # Loop over directories.
-    for directory in directories:
-        program_path = os.path.join(directory, program_name)
-        # Is there such a file, and is it executable?
-        if is_executable(program_path):
-            # Good -- that's the result.
-            return program_path
-    # Couldn't find it.
-    return None
-
-
-def is_executable(path):
-    """Return true if 'path' is an executable file."""
-
-    # Make sure 'path' exists.
-    if not os.path.isfile(path):
-        return 0
-    # If we have 'os.access' available, check for executability.
-    if hasattr(os, "access") and not os.access(path, os.X_OK):
-        return 0
-    # Looks OK.
-    return 1
-
-
-def starts_with(text, prefix):
-    """Return true if 'prefix' is a prefix of 'text'."""
-
-    return text[:len(prefix)] == prefix
-
-
-def ends_with(text, suffix):
-    """Return true if 'suffix' is a suffix of 'text'."""
-
-    return text[-len(suffix):] == suffix
-
-
-def add_exit_function(exit_function):
-    """Register 'exit_function' to be called when the program exits.
-
-    'exit_function' -- A callable that takes no arguments."""
-
-    # Add the exit function to the list of things to do at exit.
-    _exit_functions.append(exit_function)
-
-
-def exit(code):
-    """Exit the program.
-
-    'code' -- The exit code to return."""
-
-    global _exit_functions
-    
-    # First run exit functions.
-    map(lambda fn: fn(), _exit_functions)
-    # Exit.
-    sys.exit(code)
+def close_file_on_exec(fd):
+    """Prevent 'fd' from being inherited across 'exec'.
     
+    'fd' -- A file descriptor, or object providing a 'fileno()'
+    method.
+
+    This function has no effect on Windows."""
+
+    if sys.platform != "win32":
+        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
+        try:
+            flags |= fcntl.FD_CLOEXEC
+        except AttributeError:
+            # The Python 2.2 RPM shipped with Red Hat Linux 7.3 does
+            # not define FD_CLOEXEC.  Fortunately, FD_CLOEXEC is 1 on
+            # every UNIX system.
+            flags |= 1
+        fcntl.fcntl(fd, fcntl.F_SETFD, flags)
+
 
 def copy(object):
     """Make a best-effort attempt to copy 'object'.
@@ -974,22 +655,6 @@ def wrap_lines(text, columns=72, break_delimiter="\\", indent=""):
     return string.join(lines, "\n")
 
 
-def print_message(min_verbose, text):
-    """Print a status message, if the verbose level is high enough.
-
-    'min_verbose' -- The minimum verbose level for which this message
-    will be printed.  A message with minimum verbose level of zero is
-    always printed; such messages should be used only in exceptional
-    situations, and certainly never in situations likely to be "batch
-    mode" use cases.
-
-    'text' -- The text of the message."""
-
-    if verbose >= min_verbose:
-        sys.stdout.write(text)
-        sys.stdout.flush()
-
-
 def format_time(time_secs, local_time_zone=1):
     """Generate a text format representing a date and time.
 
@@ -1016,66 +681,20 @@ def format_time(time_secs, local_time_zone=1):
            "%(hour)02d:%(minute)02d %(time_zone)s" % locals()
 
 
-def shuffle(items, generator=None):
-    """Randomize the order of elements in place.
+def format_time_iso(time_secs):
+    """Generate a ISO8601-compliant formatted date and time.
 
-    'items' -- A list, or other mutable sequence.
+    The output is in the format "YYYY-MM-DDThh:mm:ss+TZ", where TZ is
+    a timezone specifier.  We always normalize to UTC (and hence
+    always use the special timezone specifier "Z"), to get proper
+    sorting behaviour.
 
-    'generator' -- The random number generator, implementing Python's
-    random number generator, to use.  If 'None', use the Python 'random'
-    module. 
+    'time_secs' -- The time to be formatted, as returned by
+    e.g. 'time.time()'.
 
-    The elements of 'items' are rearranged randomly."""
+    returns -- The formatted time as a string."""
 
-    if generator is None:
-        generator = random
-
-    length = len(items)
-    # For each element in 'items', swap the element with a
-    # randomly-chosen element that appears after it.
-    for i in range(0, length - 3):
-        j = generator.randint(i + 1, length - 1)
-        swap = items[j]
-        items[j] = items[i]
-        items[i] = swap
-
-
-def sequence_difference(seq1, seq2):
-    """Returns a sequence of items in 'seq1' but not in 'seq2'.
-
-    'seq1', 'seq2' -- Sequences."""
-
-    result = []
-    for item in seq1:
-        if item not in seq2:
-            result.append(item)
-    return result
-
-
-def make_map_from_list(list, key_function=str):
-    """Construct a map from a list.
-
-    Each element in 'list' is added to the map.  'key_function' is
-    invoked on each element to construct its key; the value is the
-    element itself.  If more than one element produces a given key, the
-    last one will appear under that key; the others will not appear."""
-
-    map = {}
-    for element in list:
-        map[key_function(element)] = element
-    return map
-
-
-def purge_keys(map, keys):
-    """Remove each entry from 'map' that matches a key in 'keys'.
-
-    'map' -- A map object.
-
-    'keys' -- A sequence of keys."""
-
-    for key in keys:
-        if map.has_key(key):
-            del map[key]
+    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(time_secs))
 
 
 def make_unique_tag():
@@ -1102,6 +721,27 @@ def split_argument_list(command):
     return argument_list
 
 
+def parse_boolean(value):
+    """Parse a boolean string.
+
+    'value' -- A string.
+
+    returns -- True if 'value' is a true string, false if 'value' is a
+    false string.
+
+    raises -- 'ValueError' if 'value' is neither a true string, nor a
+    false string."""
+
+    value = value.lower()
+    if value in ("1", "true", "yes", "on"):
+        return 1
+    elif value in ("0", "false", "no", "off"):
+        return 0
+    else:
+        raise ValueError, value
+
+    
+    
 # No 'time.strptime' on non-UNIX systems, so use this instead.  This
 # version is more forgiving, anyway, and uses our standardized timestamp
 # format. 
@@ -1192,38 +832,6 @@ def parse_time(time_string, default_local_time_zone=1):
         return int(time.mktime(time_tuple))
     
 
-import calendar
-
-if "timegm" in dir(calendar):
-    # Use timegm from the Python library.
-    from calendar import timegm
-
-else:
-    # Note: This function can be removed if we migrate permanently to a
-    # newer Python version which includes 'timegm'.
-
-    # This function is borrowed from the Python 1.6.1 distribution's
-    # 'calendar' module.  According to the Python 1.5.2 library
-    # documentation, it should be present in that version too, but it's
-    # not.
-    def timegm(tuple):
-        from calendar import *
-        EPOCH = 1970
-        year, month, day, hour, minute, second = tuple[:6]
-        assert year >= EPOCH
-        assert 1 <= month <= 12
-        days = 365*(year-EPOCH) + leapdays(EPOCH, year)
-        for i in range(1, month):
-            days = days + mdays[i]
-        if month > 2 and isleap(year):
-            days = days + 1
-        days = days + day - 1
-        hours = days*24 + hour
-        minutes = hours*60 + minute
-        seconds = minutes*60 + second
-        return seconds
-
-
 def parse_assignment(assignment):
     """Parse an 'assignment' of the form 'name=value'.
 
@@ -1265,7 +873,7 @@ def read_assignments(file):
     lines = map(string.strip, lines)
     # Drop any lines that are completely blank or lines that are
     # comments.
-    lines = filter(lambda x: x != "" and not starts_with(x, "#"),
+    lines = filter(lambda x: x != "" and not x.startswith("#"),
                    lines)
     # Go through each of the lines to process the context assignment.
     for line in lines:
@@ -1280,21 +888,15 @@ def read_assignments(file):
 # variables
 ########################################################################
 
-verbose = 0
-"""The verbosity level.  Higher values specify more output."""
+is_installed = 1
+"""True if this application has been installed.
+
+True if the application has been installed.  False if the application is
+running out of the source tree."""
 
 rc = RcConfiguration()
 """The configuration stored in system and user rc files."""
 
-# Functions that should be called when the program exits.  Each element
-# is a callable that takes no arguments.  The return value is ignored.
-_exit_functions = []
-
-# Foreign exit functions that were installed in 'sys.exitfunc' by non-QM
-# code.  These should be called after the QM-installed exit functions
-# are processed.
-_foreign_exit_functions = []
-
 # The next number to be used when handing out unqiue tag strings.
 _unique_tag = 0
 
