@@ -7,79 +7,33 @@
 # Contents:
 #   Test classes for testing command-line programs.
 #
-# Copyright (c) 2001, 2002 by CodeSourcery, LLC.  All rights reserved. 
+# Copyright (c) 2001, 2002, 2003 by CodeSourcery, LLC.  All rights reserved. 
 #
 # For license terms see the file COPYING.
 #
 ########################################################################
 
 ########################################################################
-# imports
+# Imports
 ########################################################################
 
 import cPickle
 import errno
 import os
 import qm.common
+import qm.executable
 import qm.fields
 import qm.test.base
-from   qm.test.result import *
-from   qm.test.test import *
+from   qm.test.test import Test
+from   qm.test.result import Result
 import string
 import sys
-from   threading import *
+import types
 
 ########################################################################
-# classes
+# Classes
 ########################################################################
 
-class ReadThread(Thread):
-    """An 'ReadThread' is a thread that reads from a file."""
-
-    def __init__(self, f):
-        """Construct a new 'ReadThread'.
-
-        'f' -- The file object from which to read."""
-
-        Thread.__init__(self, None, None, None)
-
-        self.f = f
-        
-            
-    def run(self):
-        """Read the data from the stream."""
-
-        try:
-            self.data = self.f.read()
-        except:
-            self.data = ""
-	self.f.close()
-
-        
-
-class WriteThread(Thread):
-    """A 'WriteThread' is a thread that writes to a file."""
-
-    def __init__(self, f, data):
-        """Construct a new 'WriteThread'.
-
-        'f' -- The file object to which to write.
-
-        'data' -- The string to be written to the file."""
-
-        Thread.__init__(self, None, None, None)
-
-        self.f = f
-        self.data = data
-        
-
-    def run(self):
-        """Write the data to the stream."""
-        
-        self.f.write(self.data)
-	self.f.close()
-
-        
 class ExecTestBase(Test):
     """Check a program's output and exit code.
 
@@ -125,7 +79,10 @@ class ExecTestBase(Test):
             of the environment variable is the name of the context
             property, prefixed with 'QMV_'.  For example, if the value
             of the context property named 'target' is available in the
-            environment variable 'QMV_target'.""" )),
+            environment variable 'QMV_target'.  Any dots in the context
+            key are replaced by a double-underscore; e.g.,
+            "CompilerTable.c_path" will become
+            "QMV_CompilerTable__c_path".""" )),
         
         qm.fields.IntegerField(
             name="exit_code",
@@ -133,10 +90,7 @@ class ExecTestBase(Test):
             description="""The expected exit code.
 
             Most programs use a zero exit code to indicate success and a
-            non-zero exit code to indicate failure.  Under Windows,
-            QMTest does not accurately report the exit code of the
-            program; all programs are treated as if they exited with
-            code zero."""
+            non-zero exit code to indicate failure."""
             ),
 
         qm.fields.TextField(
@@ -159,7 +113,19 @@ class ExecTestBase(Test):
 
             If the output written by the program does not match this
             value, the test will fail."""
-            ) 
+            ),
+
+        qm.fields.IntegerField(
+            name="timeout",
+            title="Timeout",
+            description="""The number of seconds the child program will run.
+
+            If this field is non-negative, it indicates the number of
+            seconds the child program will be permitted to run.  If this
+            field is not present, or negative, the child program will be
+            permitted to run for ever.""",
+            default_value = -1,
+            ),
         ]
 
 
@@ -171,8 +137,9 @@ class ExecTestBase(Test):
         environment = os.environ.copy()
         # Copy context variables into the environment.
         for key, value in context.items():
-            name = "QMV_" + key
-            environment[name] = value
+            if  isinstance(value, str):
+                name = "QMV_" + key.replace(".", "__")
+                environment[name] = value
         # Extract additional environment variable assignments from the
         # 'Environment' field.
         for assignment in self.environment:
@@ -205,226 +172,88 @@ class ExecTestBase(Test):
 
         # Construct the environment.
         environment = self.MakeEnvironment(context)
+        # Create the executable.
+        if self.timeout >= 0:
+            timeout = self.timeout
+        else:
+            # If no timeout was specified, we sill run this process in a
+            # separate process group and kill the entire process group
+            # when the child is done executing.  That means that
+            # orphaned child processes created by the test will be
+            # cleaned up.
+            timeout = -2
+        e = qm.executable.Filter(self.stdin, timeout)
+        # Run it.
+        exit_status = e.Run(arguments, environment, path = program)
 
-        stdin_r = None
-        stdin_w = None
-        stdout_r = None
-        stdout_w = None
-        stderr_r = None
-        stderr_w = None
-        result_r = None
-        result_w = None
-        stdin_f = None
-        stdout_f = None
-        stderr_f = None
-        result_f = None
-        
-        # Try block to clean up temporary files and file descriptors in
-        # any eventuality.
-        try:
-            # Under Windows, use popen to create the child.  It would be
-            # better to use spawn, but it is not easy to connect the
-            # standard streams for the child that way.
-            if sys.platform == "win32":
-                # Join the program and the arguments into a single
-                # command.
-                command = program + ' ' + string.join(arguments[1:], ' ')
-                # Start the command.
-                stdin_f, stdout_f, stderr_f = os.popen3(command)
-            # Under UNIX, use fork/exec to create the child.
+        # If the process terminated normally, check the outputs.
+        if sys.platform == "win32" or os.WIFEXITED(exit_status):
+            # There are no causes of failure yet.
+            causes = []
+            # The target program terminated normally.  Extract the
+            # exit code, if this test checks it.
+            if self.exit_code is None:
+                exit_code = None
+            elif sys.platform == "win32":
+                exit_code = exit_status
             else:
-                # Create pipes for all of the standard streams.
-                stdin_r, stdin_w = os.pipe()
-                stdout_r, stdout_w = os.pipe()
-                stderr_r, stderr_w = os.pipe()
+                exit_code = os.WEXITSTATUS(exit_status)
+            # Get the output generated by the program.
+            stdout = e.stdout
+            stderr = e.stderr
+            # Record the results.
+            result["ExecTest.exit_code"] = str(exit_code)
+            result["ExecTest.stdout"] = "<pre>" + stdout + "</pre>"
+            result["ExecTest.stderr"] = "<pre>" + stderr + "</pre>"
+            # Check to see if the exit code matches.
+            if exit_code != self.exit_code:
+                causes.append("exit_code")
+                result["ExecTest.expected_exit_code"] \
+                    = str(self.exit_code)
+            # Check to see if the standard output matches.
+            if not self.__CompareText(stdout, self.stdout):
+                causes.append("standard output")
+                result["ExecTest.expected_stdout"] \
+                    = "<pre>" + self.stdout + "</pre>"
+            # Check to see that the standard error matches.
+            if not self.__CompareText(stderr, self.stderr):
+                causes.append("standard error")
+                result["ExecTest.expected_stderr"] \
+                    = "<pre>" + self.stderr + "</pre>"
+            # If anything went wrong, the test failed.
+            if causes:
+                result.Fail("Unexpected %s." % string.join(causes, ", ")) 
+        elif os.WIFSIGNALED(exit_status):
+            # The target program terminated with a signal.  Construe
+            # that as a test failure.
+            signal_number = str(os.WTERMSIG(exit_status))
+            result.Fail("Program terminated by signal.")
+            result["ExecTest.signal_number"] = signal_number
+        elif os.WIFSTOPPED(exit_status):
+            # The target program was stopped.  Construe that as a
+            # test failure.
+            signal_number = str(os.WSTOPSIG(exit_status))
+            result.Fail("Program stopped by signal.")
+            result["ExecTest.signal_number"] = signal_number
+        else:
+            # The target program terminated abnormally in some other
+            # manner.  (This shouldn't normally happen...)
+            result.Fail("Program did not terminate normally.")
 
-                # Create a pipe for communicating with the child process.
-                # This pipe is used to communicate test results from the
-                # child to the parent in case something goes wrong (for
-                # instance, if the target program cannot be run).
-                #
-                # Only the parent reads from the pipe, and only the child
-                # writes to the pipe.  If the child process runs the target
-                # program successfully, it simply closes the pipe without
-                # writing anything.  If something goes wrong, though, the
-                # child process builds an appropriate test result object,
-                # pickles it, writes it to the pipe, and then closes the
-                # pipe and exits.
-                result_r, result_w = os.pipe()
 
-                # Fork a new process.
-                child_pid = os.fork()
+    def __CompareText(self, s1, s2):
+        """Compare 's1' and 's2', ignoring line endings.
 
-                if child_pid == 0:
-                    # This is the child process.
-                    try:
-                        # Close the pipe ends we do not need.
-                        os.close(stdin_w)
-                        os.close(stdout_r)
-                        os.close(stderr_r)
-                        os.close(result_r)
-                        # Redirect stdin from the standard input file.
-                        os.dup2(stdin_r, sys.stdin.fileno())
-                        # Redirect stdout to the standard output file.
-                        os.dup2(stdout_w, sys.stdout.fileno())
-                        # Redirect stderr to the standard error file.
-                        os.dup2(stderr_w, sys.stderr.fileno())
-                        # Execute the program.
-                        os.execvpe(program, arguments, environment)
-                    except:
-                        # Perhaps something went wrong while setting up
-                        # the standard stream files, or we were unable
-                        # to execute the program.
-                        exc_info = sys.exc_info()
-                        result = (Result.ERROR,
-                                  { Result.CAUSE : "Could not execute program",
-                                    Result.EXCEPTION : "%s: %s" % exc_info[:2],
-                                    Result.TRACEBACK :
-                                      qm.format_traceback(exc_info) })
-                        cPickle.dump(result, os.fdopen(result_w, "w"))
-                        # Exit.
-                        os._exit(1)
-                    else:
-                        # We should never get here.  If the call to
-                        # execve fails, an exception will be thrown.
-                        assert 0
+        's1' -- A string.
 
-                # This is the parent process.  Close file descriptors
-                # we do not need.
-                os.close(stdin_r)
-                stdin_r = None
-                os.close(stdout_w)
-                stdout_w = None
-                os.close(stderr_w)
-                stderr_w = None
-                os.close(result_w)
-                result_w = None
+        's2' -- A string.
 
-                # Create the file objects.
-                stdin_f = os.fdopen(stdin_w, "w")
-                stdin_w = None
-                stdout_f = os.fdopen(stdout_r)
-                stdout_r = None
-                stderr_f = os.fdopen(stderr_r)
-                stderr_r = None
-                
-            # Create a thread to write to the child's standard input
-            # stream.
-            stdin_thread = WriteThread(stdin_f, self.stdin)
-	    stdin_f = None
-            # Create threads to read the child's standard output and
-            # standard error streams.
-            stdout_thread = ReadThread(stdout_f)
-	    stdout_f = None
-            stderr_thread = ReadThread(stderr_f)
-	    stderr_f = None
+        returns -- True if 's1' and 's2' are the same, ignoring
+        differences in line endings."""
 
-            # Start the threads.
-            stdin_thread.start()
-            stdout_thread.start()
-            stderr_thread.start()
-            
-            # Wait for the child process to complete.
-            if sys.platform == "win32":
-                # On Windows, we have no way of obtaining the exit code.
-                exit_status = 0
-            else:
-                exit_status = os.waitpid(child_pid, 0)[1]
-            # Join the threads, so that the data read is known
-            # to be available.
-            stdin_thread.join()
-            stdout_thread.join()
-            stderr_thread.join()
-
-            pickle = None
-            
-            if sys.platform != "win32":
-                # Try to read a pickled result from the result pipe.  If the
-                # child process didn't write one, we'll read zero bytes.
-                result_f = os.fdopen(result_r)
-                result_r = None
-                pickle = result_f.read()
-                # If we read anything, there was a failure.
-                if pickle:
-                    (outcome, annotations) = cPickle.loads(pickle)
-                    result.SetOutcome(outcome)
-                    for k in annotations.keys():                    
-                        result[k] = annotations[k]
-
-            # If there was a pickle, there is nothing more to do.
-            if pickle:
-                pass
-            # If the process terminated normally, check the outputs.
-            elif sys.platform == "win32" or os.WIFEXITED(exit_status):
-                # There are no causes of failure yet.
-                causes = []
-                # The target program terminated normally.  Extract the
-                # exit code, if this test checks it.
-                if self.exit_code is None:
-                    exit_code = None
-		elif sys.platform == "win32":
-		    exit_code = 0
-                else:
-                    exit_code = os.WEXITSTATUS(exit_status)
-                # Get the output generated by the program.
-                stdout = stdout_thread.data
-                stderr = stderr_thread.data
-                # Check to see if the exit code matches.
-                if exit_code != self.exit_code:
-                    causes.append("exit_code")
-                    result["ExecTest.expected_exit_code"] \
-                        = str(self.exit_code)
-                    result["ExecTest.exit_code"] = str(exit_code)
-                # Check to see if the standard output matches.
-                if stdout != self.stdout:
-                    causes.append("standard output")
-                    result["ExecTest.stdout"] = "'''" + stdout + "'''"
-                    result["ExecTest.expected_stdout"] \
-                        = "'''" + self.stdout + "'''"
-                # Check to see that the standard error matches.
-                if stderr != self.stderr:
-                    causes.append("standard error")
-                    result["ExecTest.stderr"] = "'''" + stderr + "'''"
-                    result["ExecTest.expected_stderr"] \
-                        = "'''" + self.stderr + "'''"
-                # If anything went wrong, the test failed.
-                if causes:
-                    result.Fail("Unexpected %s." % string.join(causes, ", ")) 
-            elif os.WIFSIGNALED(exit_status):
-                # The target program terminated with a signal.  Construe
-                # that as a test failure.
-                signal_number = str(os.WTERMSIG(exit_status))
-                result.Fail("Program terminated by signal.")
-                result["ExecTest.signal_number"] = signal_number
-            elif os.WIFSTOPPED(exit_status):
-                # The target program was stopped.  Construe that as a
-                # test failure.
-                signal_number = str(os.WSTOPSIG(exit_status))
-                result.Fail("Program stopped by signal.")
-                result["ExecTest.signal_number"] = signal_number
-            else:
-                # The target program terminated abnormally in some other
-                # manner.  (This shouldn't normally happen...)
-                result.Fail("Program did not terminate normally.")
-        except:
-            result.NoteException()
-            
-        # Make sure all of the file descriptors we opened are closed.
-        for fd in (stdin_r, stdin_w, stdout_r, stdout_w,
-                   stderr_r, stderr_w, result_r, result_w):
-            if fd is not None:
-                try:
-                    os.close(fd)
-                except:
-                    pass
-
-        # Make sure all of the stream objects are closed, too.
-        for f in (stdin_f, stdout_f, stderr_f, result_f):
-            if f is not None:
-                try:
-                    f.close()
-                except:
-                    pass
+        # The "splitlines" method works independently of the line ending
+        # convention in use.
+        return s1.splitlines() == s2.splitlines()
 
         
     
@@ -457,6 +286,8 @@ class ExecTest(ExecTestBase):
             automatically."""
             ))]
 
+    _allow_arg_names_matching_class_vars = 1
+
 
     def Run(self, context, result):
         """Run the test.
@@ -475,7 +306,7 @@ class ExecTest(ExecTestBase):
             return
 
         self.RunProgram(self.program, 
-			[ self.program ] + self.arguments,
+                        [ self.program ] + self.arguments,
                         context, result)
 
 
@@ -593,7 +424,9 @@ class ShellScriptTest(ExecTestBase):
             ))        
         ]
 
+    _allow_arg_names_matching_class_vars = 1
 
+    
     def Run(self, context, result):
         """Run the test.
 
@@ -605,8 +438,15 @@ class ShellScriptTest(ExecTestBase):
         modified by this method to indicate outcomes other than
         'Result.PASS' or to add annotations."""
 
+        # On Windows, batch files must end with a ".bat" suffix or the
+        # command shell will not execute them.
+        if sys.platform == "win32":
+            suffix = ".bat"
+        else:
+            suffix = ""
         # Create a temporary file for the script.
-        self.__script_file_name, script_file = qm.open_temporary_file() 
+        self.__script_file_name, script_file \
+            = qm.open_temporary_file("w+", suffix) 
         try:
             # Write the script to the temporary file.
             script_file.write(self.script)
